# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from angst import __version__ as angst_version

project = "angst"
copyright = "2023, Nicolas Tessore"
author = "Nicolas Tessore"
version = angst_version
release = angst_version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinxcontrib.katex",
    "numpydoc",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_theme_options = {
    "light_logo": "logo-light.svg",
    "dark_logo": "logo-dark.svg",
}


# -- Autodoc -----------------------------------------------------------------

autodoc_typehints = "none"



"""angst -- angular statistics"""

__all__ = [
    "displace",
    "displacement",
    "__version__",
    "__version_tuple__",
]

from ._points import displace, displacement
from ._version import __version__, __version_tuple__



import numpy as np
import numpy.testing as npt


def test_displace_arg_complex():
    from angst import displace

    d = 5.0
    r = np.radians(d)

    # north
    lon, lat = displace(0.0, 0.0, r + 0j)
    assert np.allclose([lon, lat], [0.0, d])

    # south
    lon, lat = displace(0.0, 0.0, -r + 0j)
    assert np.allclose([lon, lat], [0.0, -d])

    # east
    lon, lat = displace(0.0, 0.0, 1j * r)
    assert np.allclose([lon, lat], [-d, 0.0])

    # west
    lon, lat = displace(0.0, 0.0, -1j * r)
    assert np.allclose([lon, lat], [d, 0.0])


def test_displace_arg_real():
    from angst import displace

    d = 5.0
    r = np.radians(d)

    # north
    lon, lat = displace(0.0, 0.0, [r, 0])
    assert np.allclose([lon, lat], [0.0, d])

    # south
    lon, lat = displace(0.0, 0.0, [-r, 0])
    assert np.allclose([lon, lat], [0.0, -d])

    # east
    lon, lat = displace(0.0, 0.0, [0, r])
    assert np.allclose([lon, lat], [-d, 0.0])

    # west
    lon, lat = displace(0.0, 0.0, [0, -r])
    assert np.allclose([lon, lat], [d, 0.0])


def test_displace_abs(rng):
    from angst import displace

    n = 1000
    abs_alpha = rng.uniform(0, 2 * np.pi, size=n)
    arg_alpha = rng.uniform(-np.pi, np.pi, size=n)

    lon_ = np.degrees(rng.uniform(-np.pi, np.pi, size=n))
    lat_ = np.degrees(np.arcsin(rng.uniform(-1, 1, size=n)))

    lon, lat = displace(lon_, lat_, abs_alpha * np.exp(1j * arg_alpha))

    th = np.radians(90.0 - lat)
    th_ = np.radians(90.0 - lat_)
    delt = np.radians(lon - lon_)

    cos_a = np.cos(th) * np.cos(th_) + np.cos(delt) * np.sin(th) * np.sin(th_)

    npt.assert_allclose(cos_a, np.cos(abs_alpha))



"""Functions to handle points on the sphere."""

from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike


def displace(
    lon: ArrayLike, lat: ArrayLike, alpha: ArrayLike
) -> tuple[ArrayLike, ArrayLike]:
    """
    Displace points with longitude *lon* and latitude *lat* (both in
    degrees) by *alpha*.  Inputs must be array-like.  The displacement
    *alpha* must be complex-valued or have a leading axis of size 2 for
    its real and imaginary component.

    Returns a tuple *newlon, newlat* of longitude and latitude (both in
    degrees) of the displaced points.
    """

    alpha = np.asanyarray(alpha)
    if np.iscomplexobj(alpha):
        alpha1, alpha2 = alpha.real, alpha.imag
    else:
        alpha1, alpha2 = alpha

    # we know great-circle navigation:
    # θ' = arctan2(√[(cosθ sin|α| - sinθ cos|α| cosγ)² + (sinθ sinγ)²],
    #              cosθ cos|α| + sinθ sin|α| cosγ)
    # δ = arctan2(sin|α| sinγ, sinθ cos|α| - cosθ sin|α| cosγ)

    t = np.radians(lat)
    ct, st = np.sin(t), np.cos(t)  # sin and cos flipped: lat not co-lat

    a = np.hypot(alpha1, alpha2)  # abs(alpha)
    g = np.arctan2(alpha2, alpha1)  # arg(alpha)
    ca, sa = np.cos(a), np.sin(a)
    cg, sg = np.cos(g), np.sin(g)

    # flipped atan2 arguments for lat instead of co-lat
    tp = np.arctan2(ct * ca + st * sa * cg, np.hypot(ct * sa - st * ca * cg, st * sg))

    d = np.arctan2(sa * sg, st * ca - ct * sa * cg)

    return lon - np.degrees(d), np.degrees(tp)


def displacement(
    from_lon: ArrayLike,
    from_lat: ArrayLike,
    to_lon: ArrayLike,
    to_lat: ArrayLike,
) -> ArrayLike:
    """
    Compute the complex displacement that transforms points with
    longitude *from_lon* and latitude *from_lat* into points with
    longitude *to_lon* and latitude *to_lat* (all in degrees).  All
    inputs must be array-like.
    """

    a = np.radians(np.subtract(90.0, to_lat))
    b = np.radians(np.subtract(90.0, from_lat))
    g = np.radians(np.subtract(from_lon, to_lon))

    sa, ca = np.sin(a), np.cos(a)
    sb, cb = np.sin(b), np.cos(b)
    sg, cg = np.sin(g), np.cos(g)

    r = np.arctan2(np.hypot(sa * cb - ca * sb * cg, sb * sg), ca * cb + sa * sb * cg)
    x = sb * ca - cb * sa * cg
    y = sa * sg
    z = np.hypot(x, y)
    return r * (x / z + 1j * y / z)  # type: ignore
